import java.awt.*;
import java.util.Random;
import java.util.ArrayList;

public final class DeansAI extends AI
{
	/* Internal Class Definitions */
	private final class TargetWeighter
	{	
		private float[][] ItemVals;
		private float[]   Weights;
		private float[]   MaxVals;
	
		public TargetWeighter (int TotalItems)
		{
			Weights  = new float[TOTAL_WEIGHTS];
			MaxVals  = new float[TOTAL_WEIGHTS];
			ItemVals = new float[TotalItems][TOTAL_WEIGHTS];
		}
		
		public void SetValue (WEIGHTPARAMS Param, int ItemIndex, float Val)
		{
			ItemVals[ItemIndex][Param.ordinal()] = Val;
		}
		
		public void SetWeight (WEIGHTPARAMS Param, float Weight)
		{
			Weights[Param.ordinal()] = Weight;
		}
		
		public int GetBestTargetIndex ()
		{
			float   BestVal   = Float.MAX_VALUE;
			int     BestIndex = NO_TARGET_FOUND;
			
			for (int i = 0; i < ItemVals.length; i++)
			{
				for (int j = 0; j < Weights.length; j++)
				{
					if (MaxVals[j] < ItemVals[i][j])
						MaxVals[j] = ItemVals[i][j];
				}
			}
			
			for (int i = 0; i < ItemVals.length; i++)
			{
				float CurrItemWeightedVal = 0;
			
				for (int j = 0; j < Weights.length; j++)
					CurrItemWeightedVal += ((ItemVals[i][j] / MaxVals[j]) * Weights[j]);
					
				if ((CurrItemWeightedVal < BestVal) && (CurrItemWeightedVal > 0))
				{
					BestVal   = CurrItemWeightedVal;
					BestIndex = i;
				}
			}

			return BestIndex;
		}
	}
	
	private class VectorAverager
	{
		private IMVector[] Samples;
		private int        CurrSample;

		public VectorAverager (int NumSamples)
		{
			Samples    = new IMVector[NumSamples];
			CurrSample = 0;
		}

		public void Flush ()
		{
			for (int i = 0; i < Samples.length; i++)
				Samples[i] = null;
		}

		public void AddNewSample (IMVector NewSample)
		{
			Samples[CurrSample] = NewSample;

			CurrSample++;

			if (CurrSample == Samples.length)
				CurrSample = 0;
		}

		public Vector GetAverage ()
		{
			int    ValidSamples = 0;
			Vector AvResult     = new Vector(0, 0);

			for (int i = 0; i < Samples.length; i++)
			{
				if (Samples[i] == null)
					break;

				AvResult.Add(Samples[i]);
				ValidSamples++;
			}

			AvResult.Divide(ValidSamples);
			return AvResult;
		}

		public IMVector GetLastSample ()
		{
			IMVector LastSample;

			if (CurrSample == 0)
				LastSample = Samples[Samples.length - 1];
			else
				LastSample = Samples[CurrSample - 1];

			if (LastSample == null)
				return new IMVector(0, 0);
			else
				return LastSample;
		}
	}

	/* Enum Definitions */
	private static enum WEIGHTPARAMS {DIST, HEALTH, ANGLEDELTA};
	private static enum STATES       {SEEK, DESTROY};

	/*Internal Utility Class Constants */
	private static final int       TOTAL_WEIGHTS            = WEIGHTPARAMS.values().length;
	private static final int       NO_TARGET_FOUND          = Integer.MAX_VALUE;

	/* AI Configuration Constants */
	private static final boolean   AI_DEBUG_ON              = false;
	
	private static final int       TARGET_SHOOT_RANGE       = 200;
	private static final float     TARGET_DIST_WEIGHT       = 1.00f;
	private static final float     TARGET_HEALTH_WEIGHT     = 0.10f;
	private static final float     TARGET_ANGLEDELTA_WEIGHT = 0.01f;
	private static final float     TARGET_CONCENTRATION     = 0.94f;

	/* AI State Variables */
	private int             TargetShipID;
	private boolean         FoundTarget;
	private STATES          CurrState                = STATES.SEEK;

	private IMVector        LastTargetPos;

	private VectorAverager  TargetVelAverage         = new VectorAverager(10);
	private VectorAverager  TargetAccAverage         = new VectorAverager(10);

	private static Random   RandGen                  = new Random(System.currentTimeMillis());


	/* Public AI Routines */
	public String GetName()
	{
		return "HAL 2000";
	}

	public String GetCreator()
	{
		return "Dean Camera";
	}

	public String GetInfo()
	{
		return "S:" + CurrState;
	}

	public void DoAI(ArrayList<IMShip> ShipList, ArrayList<IMBullet> BulletList,
	                 ArrayList<IMPlanet> PlanetList, IMShipExt Me)
	{
		if (FoundTarget)
		{
			if (CheckIfTargetExists(ShipList) == false)
			{
				DebugPrint(Me.GetShipID() + " AI:  Target " + TargetShipID +
						   " no longer exists, finding new target");

				FoundTarget = false;
			}
		}

		if (RandGen.nextFloat() > TARGET_CONCENTRATION)
		{
			DebugPrint(Me.GetShipID() + " AI:  Selecting new target");
			FoundTarget = false;
		}

		if (!(FoundTarget))
		{
			if (!(AI_ChooseNewTarget(ShipList, BulletList, Me)))
				return;
		}
				
		IMShip Target = GetTarget(TargetShipID, ShipList);

		GetTargetAverageVelAcc(Target);
		AI_ChooseState(Target, Me);
		
		switch (CurrState)
		{
			case SEEK:
				AI_DoSeek(Target, Me);
				break;
			case DESTROY:
				AI_DoDestroy(Target, Me);
				break;
		}
	}
	
	/* Private AI Routines */
	private boolean AI_ChooseNewTarget(ArrayList<IMShip>   ShipList,
	                                   ArrayList<IMBullet> BulletList,
									   IMShipExt Me)
	{
		int MyTeamID = Me.GetTeam().GetID();
		int TargetIndex;

		TargetWeighter Weighter = new TargetWeighter(ShipList.size());
		
		Weighter.SetWeight(WEIGHTPARAMS.DIST, TARGET_DIST_WEIGHT);
		Weighter.SetWeight(WEIGHTPARAMS.HEALTH, TARGET_HEALTH_WEIGHT);
		Weighter.SetWeight(WEIGHTPARAMS.ANGLEDELTA, TARGET_ANGLEDELTA_WEIGHT);

		int i = 0;

		for (IMShip CurrShip : ShipList)
		{
			if (CurrShip.GetTeam().GetID() == MyTeamID)
			{
				i++;
				continue;
			}
			
			Weighter.SetValue(WEIGHTPARAMS.DIST, i,
			                  (float)Vector.Normal(CurrShip.GetPosVector(),
			                                       Me.GetPosVector()).GetLength());

			Weighter.SetValue(WEIGHTPARAMS.HEALTH, i,
			                  (float)CurrShip.GetHealth());

			Weighter.SetValue(WEIGHTPARAMS.ANGLEDELTA, i,
			                  (float)Math.abs(Angle.GetAngleDelta(Me.GetHeading(),
			                                                      CurrShip.GetHeading())));
			i++;
		}

		TargetIndex = Weighter.GetBestTargetIndex();
		
		if (TargetIndex != NO_TARGET_FOUND)
		{
			IMShip Target = ShipList.get(TargetIndex);
		
			TargetShipID  = Target.GetShipID();
			LastTargetPos = Target.GetPosVector();

			TargetVelAverage.Flush();
			TargetAccAverage.Flush();

			DebugPrint(Me.GetShipID() + " AI:  Chosen target " + TargetShipID);
			FoundTarget = true;

			return true;
		}

		return false;
	}

	private void AI_ChooseState(IMShip Target, IMShipExt Me)
	{
		IMVector CurrPos   = Me.GetPosVector();
		IMVector TargetPos = Target.GetPosVector();
		
		if (Vector.Subtract(TargetPos, CurrPos).GetLength() <= TARGET_SHOOT_RANGE)
			CurrState = STATES.DESTROY;
		else
			CurrState = STATES.SEEK;
	}
	
	private void AI_DoSeek(IMShip Target, IMShipExt Me)
	{
		double AngTargetMe, dAng;
		
		IMVector TargetPos = GetTargetPositionAt(Target, 1);
		IMVector CurrPos   = Me.GetPosVector();

		AngTargetMe  = Vector.GetAngle(CurrPos, TargetPos);
		dAng         = Angle.GetAngleDelta(Me.GetHeading(), AngTargetMe);

		super.SetTurningPercentage((int)((dAng / Constants.SHIP_TURNING_RADS) * 100));
		super.SetThrustPercentage(100);
	}
	
	private void AI_DoDestroy(IMShip Target, IMShipExt Me)
	{
		double AngTargetMe, dAng;
		  
		IMVector TargetPos = Target.GetPosVector();
		IMVector CurrPos   = Me.GetPosVector();

		IMVector TargetPosAtBC = GetTargetPositionAt(Target,
		                                             (Vector.Normal(TargetPos, CurrPos).GetLength() /
		                                             Constants.BULLET_SPEED));

		AngTargetMe  = Vector.GetAngle(CurrPos, TargetPosAtBC);
		dAng         = Angle.GetAngleDelta(Me.GetHeading(), AngTargetMe);

		super.SetTurningPercentage((int)((dAng / Constants.SHIP_TURNING_RADS) * 100));

		if (Math.abs(dAng) <= (Angle.ONE_DEGREE * 2))
			super.SetFire(true);
		else
			super.SetThrustPercentage(80);
	}

	/* Private AI Utility Routines */
	private void DebugPrint(String Text)
	{
		if (AI_DEBUG_ON == true)
			System.out.println(Text);
	}

	private int GetFramesUntilCollision(IMBullet Bullet, IMShipExt Me, int FramesToCheck)
	{
		float CollisionRadius = (float)Math.sqrt((Constants.BULLET_RADIUS + Constants.SHIP_RADIUS) *
		                                         (Constants.BULLET_RADIUS + Constants.SHIP_RADIUS));

		for (int i = 0; i < FramesToCheck; i++)
		{
			IMVector F1 = GetMyPositionAt(Me, i);
			IMVector F2 = GetBulletPositionAt(Bullet, i);

			if (Vector.Normal(F1, F2).GetLength() <= CollisionRadius)
				return i;
		}

		return Integer.MAX_VALUE;
	}

	private IMVector GetMyPositionAt(IMShipExt Me, double Frames)
	{
		IMVector MyPos    = Me.GetPosVector();
		IMVector MyVel    = Me.GetVelVector();
		IMVector MyAcc    = Me.GetAccVector();

		float    HalfFramesSq = (float)((Frames * Frames) / 2);

		return new IMVector ((float)(MyPos.GetX() + (MyVel.GetX() * Frames) +
		                            (MyAcc.GetX() * HalfFramesSq)),
		                     (float)(MyPos.GetY() + (MyVel.GetY() * Frames) +
		                            (MyAcc.GetY() * HalfFramesSq)));
	}

	private IMVector GetTargetPositionAt(IMShip Target, double Frames)
	{
		IMVector TargetPos    = Target.GetPosVector();
		Vector   TargetVel    = TargetVelAverage.GetAverage();
		Vector   TargetAcc    = TargetAccAverage.GetAverage();

		float    HalfFramesSq = (float)((Frames * Frames) / 2);

		return new IMVector ((float)(TargetPos.GetX() + (TargetVel.GetX() * Frames) +
		                            (TargetAcc.GetX() * HalfFramesSq)),
		                     (float)(TargetPos.GetY() + (TargetVel.GetY() * Frames) +
		                            (TargetAcc.GetY() * HalfFramesSq)));
	}

	private IMVector GetBulletPositionAt(IMBullet TargetBullet, double Frames)
	{
		IMVector BulletPos = TargetBullet.GetPosVector();
		IMVector BulletVel = TargetBullet.GetVelVector();

		return new IMVector ((float)(BulletPos.GetX() + (BulletVel.GetX() * Frames)),
		                    ((float)(BulletPos.GetY() + (BulletVel.GetY() * Frames))));
	}

	private boolean CheckIfTargetExists(ArrayList<IMShip> ShipList)
	{
		boolean TargetExists = false;

		for (IMShip CurrShip : ShipList)
		{
			if (CurrShip.GetShipID() == TargetShipID)
			{
				TargetExists = true;
				break;
			}
		}

		return TargetExists;
	}

	private IMShip GetTarget(int ID, ArrayList<IMShip> ShipList)
	{
		for (IMShip CurrShip : ShipList)
		{
			if (CurrShip.GetShipID() == TargetShipID)
				return CurrShip;
		}

		return null;
	}
	
	private void GetTargetAverageVelAcc(IMShip Target)
	{
		IMVector TargetVelVector = new IMVector(Target.GetPosVector().GetX() - LastTargetPos.GetX(),
		                                        Target.GetPosVector().GetY() - LastTargetPos.GetY());
		IMVector TargetAccVector = new IMVector(TargetVelVector.GetX() -
		                                        TargetVelAverage.GetLastSample().GetX(),
		                                        TargetVelVector.GetY() -
		                                        TargetVelAverage.GetLastSample().GetY());

		TargetVelAverage.AddNewSample(TargetVelVector);
		TargetAccAverage.AddNewSample(TargetAccVector);

		LastTargetPos = Target.GetPosVector();
	}
}
